from argparse import ArgumentParser
from pathlib import Path
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import math
from typing import List, Tuple
import statsmodels.api as sm
from sklearn.metrics import mean_squared_error # is not deprecated, only squared param
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import Pipeline

TRAIN_DATA_ARG = '--train-data'
TRAIN_DATA_ARG_SHORT = '-t'
OUT_DIR_ARG = '--out'
OUT_DIR_ARG_SHORT = '-o'
DEFAULT_OUT_DIR = 'out/'
CORRELATION_THRESHOLD_ARG = '--correlation-threshold'
CORRELATION_THRESHOLD_ARG_SHORT = '-c'
DEFAULT_CORRELATION_THRESHOLD = 0.9
TARGET_ARG = '--target'
P_VALUE_THRESHOLD_ARG = '--p-value-threshold'
DEFAULT_P_VALUE_THRESHOLD = 0.05
FEATURES_ARG = '--features'
FEATURES_ARG_SHORT = '-f'
MODEL_ARG = '--model'
MODEL_ARG_SHORT = '-m'
TEST_DATA_ARG = '--test-data'
DEGREE_ARG = '--degree'
INTERACTION_ONLY_ARG = '--interaction-only'
DEGREE_BOTTOM_BOUNDARY_ARG = '--degree-bottom-boundary'
DEGREE_TOP_ARG = '--degree-top-boundary'

PVALUE_COLUMN_NAME = 'p-value'
RSQUARED_COLUMN_NAME = 'R^2'
DEGREE_COLUMN_NAME = 'degree'
INTERACTION_ONLY_COLUMN_NAME = 'interaction_only'
TRAIN_MSE_COLUMN_NAME = 'train_mse'
TEST_MSE_COLUMN_NAME = 'test_mse'

def ensure_directory(directory: Path):
    directory.mkdir(parents=True, exist_ok=True)

def get_possible_features_from_p_value(data: pd.DataFrame, target: str, p_value_threshold: float) -> pd.DataFrame:
    features = data.columns
    features = features.drop(target)
    possible_features = pd.DataFrame({
        PVALUE_COLUMN_NAME: pd.Series(dtype='float'),
        RSQUARED_COLUMN_NAME: pd.Series(dtype='float'),
    })
    for feature in features:
        pvalue, rsquared = single_feature_regression(data, feature, target)
        possible_features.loc[feature] = {PVALUE_COLUMN_NAME: pvalue, RSQUARED_COLUMN_NAME: rsquared}

    return possible_features.where(possible_features[PVALUE_COLUMN_NAME] < p_value_threshold).dropna()

def multi_feature_regression_model(train_data: pd.DataFrame, selected_features: List[str] | None, p_value_threshold: float | None, target: str) -> Tuple[sm.OLS, List[str]]:
    features: List[str] = None
    if selected_features is not None and p_value_threshold is None:
        features = selected_features
    elif p_value_threshold is not None and selected_features is None:        
        features = list(get_possible_features_from_p_value(train_data, target, p_value_threshold).index)
    else:
        raise ValueError(f'selected_features is {selected_features} and p_value_threshold is {p_value_threshold}, but expected exactly one to be set')
    
    X = sm.add_constant(train_data[features])
    y = train_data[target]
    model = sm.OLS(y, X).fit()

    return model, features

def correlation_analysis(train_data_file: Path, out_dir: Path, correlation_threshold: float):
    # Load training and test data
    train_data = pd.read_csv(train_data_file)

    # Display correlation matrix
    correlation_matrix = train_data.corr()
    plt.figure(figsize=(10, 8))
    sns.heatmap(correlation_matrix, annot=True, cmap='coolwarm', vmin=-1, vmax=1)
    plt.title('Correlation Matrix')

    ensure_directory(out_dir)
    correlation_matrix_file = out_dir / 'correlation_matrix.png'
    plt.savefig(correlation_matrix_file, bbox_inches='tight')

    # Identify highly correlated variables
    high_correlation_matrix = correlation_matrix[correlation_matrix.abs() >= correlation_threshold] # outputs a matrix, where all values < correlation_threshold are NaN and values >= correlation_threshold are their correlation
    print('High correlation matrix')
    print(high_correlation_matrix)

    high_correlations: List[Tuple[str, str, int]] = []
    for i in range(len(correlation_matrix.index)):
        for j in range(i):
            correlation_value: float = high_correlation_matrix.iloc[i, j]
            if not math.isnan(correlation_value):
                high_correlations.append((correlation_matrix.index[i], correlation_matrix.columns[j], correlation_value))
    
    if not high_correlations:
        print("No high correlations found.")
    else:
        correlations = [f'{column}/{row}: {correlation_value:.5f}' for column, row, correlation_value in high_correlations]
        print(f'High correlations found in {', '.join(correlations)}')

        ensure_directory(out_dir)
        correlations_file = out_dir / "correlations.txt"
        with open(correlations_file, 'w') as f:
            f.writelines(correlations)

def single_feature_regression(data: pd.DataFrame, feature: str, target: str) -> Tuple[float, float]:
    X = sm.add_constant(data[[feature]])  # Add constant for intercept
    y = data[target]
    model = sm.OLS(y, X).fit()
    return model.pvalues.iloc[1], model.rsquared

def best_single_feature_regression(train_data_file: Path, target: str, p_value_threshold: float, out_dir: Path):
    train_data = pd.read_csv(train_data_file)
    evaluated_features = get_possible_features_from_p_value(train_data, target, p_value_threshold)

    print('Evaluated features')
    print(evaluated_features)

    plt.figure(figsize=(1.75, 4.8))
    evaluated_pvalues = evaluated_features[[PVALUE_COLUMN_NAME]]
    sns.heatmap(evaluated_pvalues, annot=True, cmap='coolwarm', vmin=0, vmax=1)

    ensure_directory(out_dir)
    evaluated_pvalues_file_path = out_dir / "evaluated_pvalues.png"
    plt.savefig(evaluated_pvalues_file_path, bbox_inches='tight')

    plt.figure(figsize=(1.75, 4.8))
    evaluated_rsquares = evaluated_features[[RSQUARED_COLUMN_NAME]]
    sns.heatmap(evaluated_rsquares, annot=True, cmap='coolwarm', vmin=0, vmax=1)

    ensure_directory(out_dir)
    evaluated_rsquares_file_path = out_dir / "evaluated_rsquares.png"
    plt.savefig(evaluated_rsquares_file_path, bbox_inches='tight')

    possible_features = evaluated_features.where(evaluated_features[PVALUE_COLUMN_NAME] < p_value_threshold).dropna()
    possible_features = possible_features.sort_values(RSQUARED_COLUMN_NAME, ascending=False)

    print()
    print('Features matching p-value-threshold')
    print(possible_features)

    best_feature: pd.Series = None
    if possible_features.shape[0] > 0:
        best_feature = possible_features.iloc[0]
    else:
        print('No feature meets all criteria')
        return

    print()
    print('Best Feature')
    print(best_feature)

    ensure_directory(out_dir)
    best_feature_file = out_dir / 'best_feature.txt'
    with open(best_feature_file, 'w') as f:
        f.write(f'''Name: {best_feature.name}
p-value: {best_feature[PVALUE_COLUMN_NAME]}
R^2: {best_feature[RSQUARED_COLUMN_NAME]}
''')
        
def multi_feature_regression(train_data_file: Path, target: str, selected_features: List[str] | None, p_value_threshold: float | None, out_dir: Path):
    train_data = pd.read_csv(train_data_file)
    y = train_data[target]
    model, features = multi_feature_regression_model(train_data, selected_features, p_value_threshold, target)
    print(model.summary())

    ensure_directory(out_dir)
    multi_feature_regression_results_file = out_dir / 'multi_feature_regression_train_results.txt'
    with open(multi_feature_regression_results_file, 'w') as f:
        f.write(f'''features: {features}
rsquared: {model.rsquared},
train_MSE: {mean_squared_error(y, model.fittedvalues)}
''')
    
    ensure_directory(out_dir)
    model_file = out_dir / 'multi_feature_regression_model.pickle'
    model.save(model_file)

def model_evaluation(model_file: Path, test_data_file: Path, features: List[str], target: str, out_dir: Path):
    test_data = pd.read_csv(test_data_file)
    model: sm.OLS = sm.load(model_file)

    X_test = sm.add_constant(test_data[features])
    y_test = test_data[target]

    y_pred = model.predict(X_test)
    test_mse = mean_squared_error(y_test, y_pred)
    print(f'Test MSE: {test_mse}')

    ensure_directory(out_dir)
    model_evaluation_file = out_dir / 'model_evaluation.txt'
    with open(model_evaluation_file, 'w') as f:
        f.write(f'''test_MSE: {test_mse}
''')
        
def higher_order_terms(train_data_file: Path, test_data_file: Path, features: List[str], target: str, interaction_only: bool, degree: int, out_dir: Path):
    train_data = pd.read_csv(train_data_file)
    test_data = pd.read_csv(test_data_file)

    X_train = train_data[features]
    y_train = train_data[target]

    X_test = test_data[features]
    y_test = test_data[target]

    pipeline = Pipeline([
        ('poly', PolynomialFeatures(degree=degree, include_bias=False, interaction_only=interaction_only)),
        ('linear', LinearRegression())
    ])

    pipeline.fit(X_train, y_train)
    y_pred = pipeline.predict(X_train)
    train_mse = mean_squared_error(y_train, y_pred)
    rsquared = pipeline.score(X_train, y_train)

    y_pred = pipeline.predict(X_test)
    test_mse = mean_squared_error(y_test, y_pred)

    print(f'train MSE: {train_mse}, test mse: {test_mse}, R^2: {rsquared}')

    ensure_directory(out_dir)
    output_file = out_dir / f'higher_order_terms_{degree}_{interaction_only}.txt'
    with open(output_file, 'w') as f:
        f.write(f'''train MSE: {train_mse}
test MSE: {test_mse}
rsquared: {rsquared}
''')
    
    return train_mse, test_mse, rsquared
        
def best_higher_order_terms(train_data_file: Path, test_data_file: Path, features: List[str], target: str, degree_bottom_boundary: int, degree_top_boundary: int, out_dir: Path):
    if degree_bottom_boundary > degree_top_boundary:
        raise ValueError('degree bottom boundary must be smaller or equal than degree top boundary')
    
    configurations = pd.DataFrame({
        DEGREE_COLUMN_NAME: pd.Series(dtype=int),
        INTERACTION_ONLY_COLUMN_NAME: pd.Series(dtype=bool),
        TRAIN_MSE_COLUMN_NAME: pd.Series(dtype=float),
        TEST_MSE_COLUMN_NAME: pd.Series(dtype=float),
        RSQUARED_COLUMN_NAME: pd.Series(dtype='float'),
    })
    for degree in range(degree_bottom_boundary, degree_top_boundary + 1):
        interaction_only = False
        train_mse, test_mse, rsquared = higher_order_terms(train_data_file, test_data_file, features, target, interaction_only, degree, out_dir)
        configurations.loc[-1] = {
            DEGREE_COLUMN_NAME: degree,
            INTERACTION_ONLY_COLUMN_NAME: interaction_only,
            TRAIN_MSE_COLUMN_NAME: train_mse,
            TEST_MSE_COLUMN_NAME: test_mse,
            RSQUARED_COLUMN_NAME: rsquared,
        }        
        configurations.index = configurations.index + 1

        interaction_only = True
        train_mse, test_mse, rsquared = higher_order_terms(train_data_file, test_data_file, features, target, interaction_only, degree, out_dir)
        configurations.loc[-1] = {
            DEGREE_COLUMN_NAME: degree,
            INTERACTION_ONLY_COLUMN_NAME: interaction_only,
            TRAIN_MSE_COLUMN_NAME: train_mse,
            TEST_MSE_COLUMN_NAME: test_mse,
            RSQUARED_COLUMN_NAME: rsquared,
        }
        configurations.index = configurations.index + 1

    configurations = configurations.sort_values(RSQUARED_COLUMN_NAME, ascending=False)
    
    print(configurations)

def main():
    argument_parser = ArgumentParser('vqcfim', description='Virtual Quality Control for Injection Molding')
    subparsers = argument_parser.add_subparsers(title='action')

    correlation_analysis_subparser = subparsers.add_parser('correlation-analysis', aliases=['ca'], description='Gets variables with correlation coefficients >= --correlation-coefficient')
    correlation_analysis_subparser.add_argument(TRAIN_DATA_ARG, TRAIN_DATA_ARG_SHORT, action='store', type=Path, required=True)
    correlation_analysis_subparser.add_argument(OUT_DIR_ARG, OUT_DIR_ARG_SHORT, action='store', type=Path, required=False, default=DEFAULT_OUT_DIR)
    correlation_analysis_subparser.add_argument(CORRELATION_THRESHOLD_ARG, CORRELATION_THRESHOLD_ARG_SHORT, action='store', type=float, required=False, default=DEFAULT_CORRELATION_THRESHOLD)
    correlation_analysis_subparser.set_defaults(func=lambda train_data, out, correlation_threshold, func: correlation_analysis(train_data, out, correlation_threshold))

    best_single_feature_regression_subparser = subparsers.add_parser('best-single-feature-regression', aliases=['bsfr'], description='Evaluates the best single feature regression feature from the dataset')
    best_single_feature_regression_subparser.add_argument(TRAIN_DATA_ARG, TRAIN_DATA_ARG_SHORT, action='store', type=Path, required=True)
    best_single_feature_regression_subparser.add_argument(TARGET_ARG, action='store', type=str, required=True)
    best_single_feature_regression_subparser.add_argument(OUT_DIR_ARG, OUT_DIR_ARG_SHORT, action='store', type=Path, required=False, default=DEFAULT_OUT_DIR)
    best_single_feature_regression_subparser.add_argument(P_VALUE_THRESHOLD_ARG, action='store', type=float, required=False, default=DEFAULT_P_VALUE_THRESHOLD)
    best_single_feature_regression_subparser.set_defaults(func=lambda train_data, target, out, p_value_threshold, func: best_single_feature_regression(train_data, target, p_value_threshold, out))

    multi_feature_regression_subparser = subparsers.add_parser('multi-feature-regression', aliases=['mfr'], description='Performs a linear regression using multiple features')
    multi_feature_regression_subparser.add_argument(TRAIN_DATA_ARG, TRAIN_DATA_ARG_SHORT, action='store', type=Path, required=True)
    multi_feature_regression_subparser.add_argument(TARGET_ARG, action='store', type=str, required=True)
    multi_feature_regression_subparser.add_argument(OUT_DIR_ARG, OUT_DIR_ARG_SHORT, action='store', type=Path, required=False, default=DEFAULT_OUT_DIR)
    multi_feature_regression_features_group = multi_feature_regression_subparser.add_mutually_exclusive_group(required=True)
    multi_feature_regression_features_group.add_argument(FEATURES_ARG, FEATURES_ARG_SHORT, action='store', type=str, required=False, nargs='+')
    multi_feature_regression_features_group.add_argument(P_VALUE_THRESHOLD_ARG, action='store', type=float, required=False, default=None)
    multi_feature_regression_subparser.set_defaults(func=lambda train_data, target, out, features, p_value_threshold, func: multi_feature_regression(train_data, target, features, p_value_threshold, out))

    model_evaluation_subparser = subparsers.add_parser('model-evaluation', aliases=['me'], description='Evaluates a model')
    model_evaluation_subparser.add_argument(MODEL_ARG, MODEL_ARG_SHORT, action='store', type=Path, required=True)
    model_evaluation_subparser.add_argument(TEST_DATA_ARG, action='store', type=Path, required=True)
    model_evaluation_subparser.add_argument(TARGET_ARG, action='store', type=str, required=True)
    model_evaluation_subparser.add_argument(OUT_DIR_ARG, OUT_DIR_ARG_SHORT, action='store', type=Path, required=False, default=DEFAULT_OUT_DIR)
    model_evaluation_subparser.add_argument(FEATURES_ARG, FEATURES_ARG_SHORT, action='store', type=str, required=True, nargs='+')
    model_evaluation_subparser.set_defaults(func=lambda model, test_data, target, out, features, func: model_evaluation(model, test_data, features, target, out))

    higher_order_terms_subparser = subparsers.add_parser('higher-order-terms', aliases=['hot'], description='Creates and evaluates a model with higher order terms')
    higher_order_terms_subparser.add_argument(TRAIN_DATA_ARG, TRAIN_DATA_ARG_SHORT, action='store', type=Path, required=True)
    higher_order_terms_subparser.add_argument(TEST_DATA_ARG, action='store', type=Path, required=True)
    higher_order_terms_subparser.add_argument(FEATURES_ARG, FEATURES_ARG_SHORT, action='store', type=str, required=True, nargs='+')
    higher_order_terms_subparser.add_argument(TARGET_ARG, action='store', type=str, required=True)
    higher_order_terms_subparser.add_argument(OUT_DIR_ARG, OUT_DIR_ARG_SHORT, action='store', type=Path, required=False, default=DEFAULT_OUT_DIR)
    higher_order_terms_subparser.add_argument(DEGREE_ARG, action='store', type=int, required=True)
    higher_order_terms_subparser.add_argument(INTERACTION_ONLY_ARG, action='store_true', required=False)
    higher_order_terms_subparser.set_defaults(func=lambda train_data, test_data, features, target, degree, interaction_only, out, func: higher_order_terms(train_data, test_data, features, target, interaction_only, degree, out))

    best_higher_order_terms_subparser = subparsers.add_parser('best-higher-order-terms', aliases=['bhot'], description='Creates and evaluates models with higher order terms')
    best_higher_order_terms_subparser.add_argument(TRAIN_DATA_ARG, TRAIN_DATA_ARG_SHORT, action='store', type=Path, required=True)
    best_higher_order_terms_subparser.add_argument(TEST_DATA_ARG, action='store', type=Path, required=True)
    best_higher_order_terms_subparser.add_argument(FEATURES_ARG, FEATURES_ARG_SHORT, action='store', type=str, required=True, nargs='+')
    best_higher_order_terms_subparser.add_argument(TARGET_ARG, action='store', type=str, required=True)
    best_higher_order_terms_subparser.add_argument(OUT_DIR_ARG, OUT_DIR_ARG_SHORT, action='store', type=Path, required=False, default=DEFAULT_OUT_DIR)
    best_higher_order_terms_subparser.add_argument(DEGREE_BOTTOM_BOUNDARY_ARG, action='store', type=int, required=True)
    best_higher_order_terms_subparser.add_argument(DEGREE_TOP_ARG, action='store', type=int, required=True)
    best_higher_order_terms_subparser.set_defaults(func=lambda train_data, test_data, features, degree_bottom_boundary, degree_top_boundary, target, out, func: best_higher_order_terms(train_data, test_data, features, target, degree_bottom_boundary, degree_top_boundary, out))

    parsed_args = argument_parser.parse_args()
    args = vars(parsed_args)
    parsed_args.func(**args)

if __name__ == "__main__":
    main()
